from bzflag.networking.game_data_player_score import PlayerScore
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.network_protocol import NetworkProtocol
from bzflag.networking.packet import Packet


class MsgAddPlayerPacket(GamePacket):
    __slots__ = (
        'player_index',
        'player_type',
        'team_value',
        'callsign',
        'motto',
        'score',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgAddPlayer'

        self.player_index: int = -1
        self.player_type: int = -1
        self.team_value: int = -1
        self.score: PlayerScore = PlayerScore()
        self.callsign: str = ''
        self.motto: str = ''

    def _unpack(self):
        self.player_index = Packet.unpack_uint8(self.buffer)
        self.player_type = Packet.unpack_uint16(self.buffer)
        self.team_value = Packet.unpack_uint16(self.buffer)
        self.score.wins = Packet.unpack_uint16(self.buffer)
        self.score.losses = Packet.unpack_uint16(self.buffer)
        self.score.team_kills = Packet.unpack_uint16(self.buffer)

        self.callsign = Packet.unpack_string(self.buffer, NetworkProtocol.CALLSIGN_LEN)
        self.motto = Packet.unpack_string(self.buffer, NetworkProtocol.MOTTO_LEN)



from bzflag.utilities.json_serializable import JsonSerializable


class ReplayDuration(JsonSerializable):
    __slots__ = [
        'as_seconds',
        'days',
        'hours',
        'minutes',
        'seconds',
        'usecs',
    ]

    def __init__(self, timestamp: int):
        """
        A developer-friendly representation of microseconds

        :param timestamp: The duration of the recording in microseconds
        """
        super().__init__()

        secs = timestamp / 1000000

        day_len = 24 * 60 * 60
        self.days = int(secs / day_len)
        secs = secs % day_len

        hour_len = 60 * 60
        self.hours = int(secs / hour_len)
        secs = secs % hour_len

        min_len = 60
        self.minutes = int(secs / min_len)
        secs = secs % min_len

        self.seconds = int(secs)
        self.usecs = int(timestamp % 1000000)

        # Short cut for accessing this duration in seconds
        self.as_seconds = (self.days * day_len) + (self.hours * hour_len) + (self.minutes * min_len) + self.seconds



from typing import List

from bzflag.networking.game_data_score import ScoreData
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgScorePacket(GamePacket):
    __slots__ = (
        'scores',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgScore'
        self.scores: List[ScoreData] = []

    def _unpack(self):
        count: int = Packet.unpack_uint8(self.buffer)

        for i in range(0, count):
            data: ScoreData = ScoreData()
            data.player_id = Packet.unpack_uint8(self.buffer)
            data.wins = Packet.unpack_uint16(self.buffer)
            data.losses = Packet.unpack_uint16(self.buffer)
            data.team_kills = Packet.unpack_uint16(self.buffer)

            self.scores.append(data)



from typing import Optional

from bzflag.networking.game_data import GameData
from bzflag.networking.game_data_shot import ShotData


class FiringInfoData(GameData):
    __slots__ = (
        'time_sent',
        'shot',
        'flag',
        'lifetime',
    )

    def __init__(self):
        super().__init__()

        self.time_sent: float = 0.0
        self.shot: Optional[ShotData] = None
        self.flag: str = ''
        self.lifetime: float = 0.0



from bzflag.networking.game_data import GameData
from bzflag.networking.network_protocol import Vector3F


class FlagData(GameData):
    __slots__ = (
        'index',
        'abbv',
        'status',
        'endurance',
        'owner',
        'position',
        'launch_pos',
        'landing_pos',
        'flight_time',
        'flight_end',
        'initial_velocity',
    )

    def __init__(self):
        super().__init__()

        self.index: int = -1
        self.abbv: str = ''
        self.status: int = -1
        self.endurance: int = -1
        self.owner: int = -1
        self.position: Vector3F = (0.0, 0.0, 0.0)
        self.launch_pos: Vector3F = (0.0, 0.0, 0.0)
        self.landing_pos: Vector3F = (0.0, 0.0, 0.0)
        self.flight_time: float = 0.0
        self.flight_end: float = 0.0
        self.initial_velocity: float = 0.0



from itertools import chain
from typing import Dict, Union, List


class JsonSerializable:
    __slots__ = (
        'json_ignored',
    )

    def __init__(self):
        self.json_ignored: List[str] = []

    def is_json_ignored(self, slot: str) -> bool:
        if slot == 'json_ignored':
            return True

        if slot in self.json_ignored:
            return True

        return False

    def to_dict(self):
        result: Dict[str, Union[list, str]] = {}
        slots = chain.from_iterable(getattr(cls, '__slots__', []) for cls in type(self).__mro__)

        for slot in slots:
            if self.is_json_ignored(slot):
                continue

            result[JsonSerializable.jsonify_name(slot)] = getattr(self, slot)

        return result

    @staticmethod
    def jsonify_name(string: str) -> str:
        raw = string.strip('_').title().replace('_', '')

        return raw[:1].lower() + raw[1:]



from typing import Optional

from bzflag.networking.game_data_flag import FlagData
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.network_message import code_from_chars
from bzflag.networking.packet import Packet


class MsgKilledPacket(GamePacket):
    __slots__ = (
        'victim_id',
        'killer_id',
        'reason',
        'shot_id',
        'flag',
        'physics_driver_id',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgKilled'
        self.victim_id: int = -1
        self.killer_id: int = -1
        self.reason: int = -1
        self.shot_id: int = -1
        self.flag: Optional[FlagData] = None
        self.physics_driver_id: int = -1

    def _unpack(self):
        self.victim_id = Packet.unpack_uint8(self.buffer)
        self.killer_id = Packet.unpack_uint8(self.buffer)
        self.reason = Packet.unpack_uint16(self.buffer)
        self.shot_id = Packet.unpack_uint16(self.buffer)
        self.flag = Packet.unpack_flag(self.buffer)

        if self.reason == code_from_chars('pd'):
            self.physics_driver_id = Packet.unpack_uint32(self.buffer)



from bzflag.networking.game_data import GameData
from bzflag.networking.network_protocol import Vector3F

DeadStatus = 0  # not alive, not paused, etc.
Alive = 1 << 0  # player is alive
Paused = 1 << 1  # player is paused
Exploding = 1 << 2  # currently blowing up
Teleporting = 1 << 3  # teleported recently
FlagActive = 1 << 4  # flag special powers active
CrossingWall = 1 << 5  # tank crossing building wall
Falling = 1 << 6  # tank accel'd by gravity
OnDriver = 1 << 7  # tank is on a physics driver
UserInputs = 1 << 8  # user speed and angvel are sent
JumpJets = 1 << 9  # tank has jump jets on
PlaySound = 1 << 10  # play one or more sounds


class PlayerStateData(GameData):
    __slots__ = (
        'position',
        'velocity',
        'azimuth',
        'angular_velocity',
        'physics_driver',
        'user_speed',
        'user_ang_vel',
        'jump_jets_scale',
        'sounds',
    )

    def __init__(self):
        super().__init__()

        self.position: Vector3F = (0.0, 0.0, 0.0)
        self.velocity: Vector3F = (0.0, 0.0, 0.0)
        self.azimuth: float = 0.0
        self.angular_velocity: float = 0.0
        self.physics_driver: int = -1
        self.user_speed: float = 0.0
        self.user_ang_vel: float = 0.0
        self.jump_jets_scale: float = 0.0
        self.sounds: int = -1



from bzflag.networking.game_data import GameData


class ScoreData(GameData):
    __slots__ = (
        'player_id',
        'wins',
        'losses',
        'team_kills',
    )

    def __init__(self):
        super().__init__()

        self.player_id: int = -1
        self.wins: int = 0
        self.losses: int = 0
        self.team_kills: int = 0



from typing import Optional

from bzflag.networking.game_data_shot import ShotData
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgGMUpdatePacket(GamePacket):
    __slots__ = (
        'target',
        'shot',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgGMUpdate'
        self.target: int = -1
        self.shot: Optional[ShotData] = None

    def _unpack(self):
        self.target = Packet.unpack_uint8(self.buffer)
        self.shot = Packet.unpack_shot(self.buffer)



import json

from typing import List

from bzflag.networking.game_packet import GamePacket
from bzflag.utilities.json_serializable import JsonSerializable


class RRLogEncoder(json.JSONEncoder):
    black_list: List[str] = []
    white_list: List[str] = []

    def default(self, obj):
        if not isinstance(obj, JsonSerializable):
            return str(obj)

        if len(RRLogEncoder.white_list) > 0:
            if isinstance(obj, GamePacket) and obj.packet_type not in RRLogEncoder.white_list:
                return {}

        if isinstance(obj, GamePacket) and obj.packet_type in RRLogEncoder.black_list:
            return {}

        return obj.to_dict()



from datetime import timedelta
from typing import BinaryIO, List

from bzflag.networking.game_packet import GamePacket
from bzflag.networking.game_packet_mapping import GamePacketMap
from bzflag.networking.packet_invalid_error import PacketInvalidError
from bzflag.utilities.json_serializable import JsonSerializable
from bzflag.networking.network_message import NetworkMessage, chars_from_code
from bzflag.networking.packet import Packet
from bzflag.replay_header import ReplayHeader


class Replay(JsonSerializable):
    __slots__ = (
        'header',
        'start_time',
        'end_time',
        'packets',
        'errors',
    )

    def __init__(self, file: str):
        super().__init__()

        self.header: ReplayHeader = ReplayHeader()
        self.packets: List[GamePacket] = []
        self.errors: List[str] = []

        with open(file, 'rb') as replay_file:
            self.header.unpack(replay_file)
            self._calc_timestamps(replay_file)
            self._load_packets(replay_file)

    def _calc_timestamps(self, buf: BinaryIO) -> None:
        # We want to load the first packet (a null packet) so we can get its
        # timestamp
        packet: Packet = Packet()
        packet.unpack(buf)

        if packet.timestamp is None:
            raise PacketInvalidError

        self.start_time = packet.timestamp
        self.end_time = packet.timestamp + timedelta(microseconds=self.header.file_time)

    def _load_packets(self, buf: BinaryIO) -> None:
        # We've loaded the replay header already, so let's save the current
        # starting position for the packets
        packets_start = buf.tell()

        # Get the full file size and save it
        buf.seek(0, 2)
        file_size = buf.tell()

        # Go back to the start of the packets so we can read them in
        buf.seek(packets_start)

        while True:
            packet = Packet()
            packet.unpack(buf)

            try:
                msg_code = NetworkMessage(packet.code)
                game_packet = GamePacketMap[msg_code]()
                game_packet.packet = packet
                game_packet.timestamp_offset = self.end_time
                game_packet.build()

                self.packets.append(game_packet)
            except KeyError:
                game_code: str = chars_from_code(packet.code)
                self.errors.append(f'Unsupported game packet code: {game_code}')

            # We've reached the end of the replay
            if buf.tell() == file_size:
                break



from bzflag.networking.game_data import GameData
from bzflag.networking.network_protocol import Vector3F


class ShotData(GameData):
    __slots__ = (
        'player_id',
        'shot_id',
        'position',
        'velocity',
        'delta_time',
        'team',
    )

    def __init__(self):
        super().__init__()

        self.player_id: int = -1
        self.shot_id: int = -1
        self.position: Vector3F = (0.0, 0.0, 0.0)
        self.velocity: Vector3F = (0.0, 0.0, 0.0)
        self.delta_time: float = 0.0
        self.team: int = -1



from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgCaptureFlagPacket(GamePacket):
    __slots__ = (
        'player_id',
        'flag_id',
        'team',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgCaptureFlag'
        self.player_id: int = -1
        self.flag_id: int = -1
        self.team: int = -1

    def _unpack(self):
        self.player_id = Packet.unpack_uint8(self.buffer)
        self.flag_id = Packet.unpack_uint16(self.buffer)
        self.team = Packet.unpack_uint16(self.buffer)



from bzflag.networking.game_packet import GamePacket


class MsgNullPacket(GamePacket):
    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgNull'

    def _unpack(self):
        pass



from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgTimeUpdatePacket(GamePacket):
    __slots__ = (
        'time_left',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgTimeUpdate'
        self.time_left: int = -1

    def _unpack(self):
        self.time_left = Packet.unpack_int32(self.buffer)



from typing import Optional

from bzflag.networking.game_data_firing_info import FiringInfoData
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgShotBeginPacket(GamePacket):
    __slots__ = (
        'firing_info',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgShotBegin'
        self.firing_info: Optional[FiringInfoData] = None

    def _unpack(self):
        self.firing_info = Packet.unpack_firing_info(self.buffer)



from typing import List

from bzflag.networking.game_data_flag import FlagData
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgFlagUpdatePacket(GamePacket):
    __slots__ = (
        'flags',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgFlagUpdate'
        self.flags: List[FlagData] = []

    def _unpack(self):
        count: int = Packet.unpack_uint16(self.buffer)

        for i in range(0, count):
            flag = Packet.unpack_flag(self.buffer)
            self.flags.append(flag)



from enum import Enum


def code_from_chars(code: str) -> int:
    return int.from_bytes(str.encode(code), byteorder='big')


def chars_from_code(chars: int) -> str:
    return chars.to_bytes(2, byteorder='big').decode('utf-8')


class NetworkMessage(Enum):
    Null = 0
    Accept = code_from_chars('ac')
    AdminInfo = code_from_chars('ai')
    Alive = code_from_chars('al')
    AddPlayer = code_from_chars('ap')
    AutoPilot = code_from_chars('au')
    CaptureFlag = code_from_chars('cf')
    CustomSound = code_from_chars('cs')
    CacheURL = code_from_chars('cu')
    DropFlag = code_from_chars('df')
    Enter = code_from_chars('en')
    Exit = code_from_chars('ex')
    FlagType = code_from_chars('ft')
    FlagUpdate = code_from_chars('fu')
    FetchResources = code_from_chars('fr')
    GrabFlag = code_from_chars('gf')
    GMUpdate = code_from_chars('gm')
    GetWorld = code_from_chars('gw')
    GameSettings = code_from_chars('gs')
    GameTime = code_from_chars('gt')
    Handicap = code_from_chars('hc')
    Killed = code_from_chars('kl')
    LagState = code_from_chars('ls')
    Message = code_from_chars('mg')
    NearFlag = code_from_chars('Nf')
    NewRabbit = code_from_chars('nR')
    NegotiateFlags = code_from_chars('nf')
    Pause = code_from_chars('pa')
    PlayerInfo = code_from_chars('pb')
    PlayerUpdate = code_from_chars('pu')
    PlayerUpdateSmall = code_from_chars('ps')
    QueryGame = code_from_chars('qg')
    QueryPlayers = code_from_chars('qp')
    Reject = code_from_chars('rj')
    RemovePlayer = code_from_chars('rp')
    ReplayReset = code_from_chars('rr')
    ShotBegin = code_from_chars('sb')
    Score = code_from_chars('sc')
    ScoreOver = code_from_chars('so')
    ShotEnd = code_from_chars('se')
    SuperKill = code_from_chars('sk')
    SetVar = code_from_chars('sv')
    TimeUpdate = code_from_chars('to')
    Teleport = code_from_chars('tp')
    TransferFlag = code_from_chars('tf')
    TeamUpdate = code_from_chars('tu')
    WantWHash = code_from_chars('wh')
    WantSettings = code_from_chars('ws')
    PortalAdd = code_from_chars('Pa')
    PortalRemove = code_from_chars('Pr')
    PortalUpdate = code_from_chars('Pu')



from typing import Optional

from bzflag.networking.game_data_player_state import PlayerStateData
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgPlayerUpdatePacket(GamePacket):
    __slots__ = (
        'player_id',
        'state',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgPlayerUpdate'
        self.player_id: int = -1
        self.state: Optional[PlayerStateData] = None

    def _unpack(self):
        # Discard this value; I'm not sure why this value comes out to a weird
        # float. We have the timestamp of the raw packet, so just that instead
        _ = Packet.unpack_float(self.buffer)

        self.player_id = Packet.unpack_uint8(self.buffer)
        self.state = Packet.unpack_player_state(self.buffer, self.packet.code)



from typing import List

from bzflag.networking.game_data_player_info import PlayerInfo
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgAdminInfoPacket(GamePacket):
    __slots__ = (
        'players',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgAdminInfo'
        self.players: List[PlayerInfo] = []

    def _unpack(self):
        count: int = Packet.unpack_uint8(self.buffer)

        for i in range(0, count):
            Packet.unpack_uint8(self.buffer)

            p_info: PlayerInfo = PlayerInfo()
            p_info.player_index = Packet.unpack_uint8(self.buffer)
            p_info.ip_address = Packet.unpack_ip_address(self.buffer)

            self.players.append(p_info)



from bzflag.networking.game_data import GameData

IsRegistered = 1 << 0
IsVerified = 1 << 1
IsAdmin = 1 << 2


class PlayerData(GameData):
    __slots__ = (
        'player_id',
        'is_registered',
        'is_verified',
        'is_admin',
    )

    def __init__(self):
        super().__init__()

        self.player_id = -1
        self.is_registered = False
        self.is_verified = False
        self.is_admin = False



import warnings

from bzflag.networking.msg_flag_drop import MsgFlagDropPacket


class MsgDropFlagPacket(MsgFlagDropPacket):
    warnings.warn("use the `MsgFlagDropPacket` class now", DeprecationWarning)

    pass



from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgMessagePacket(GamePacket):
    __slots__ = (
        'player_from_id',
        'player_to_id',
        'message',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgMessage'
        self.player_from_id: int = -1
        self.player_to_id: int = -1
        self.message: str = ''

    def _unpack(self):
        self.player_from_id = Packet.unpack_uint8(self.buffer)
        self.player_to_id = Packet.unpack_uint8(self.buffer)
        self.message = Packet.unpack_string(self.buffer, -1)



from typing import Tuple

Vector3F = Tuple[float, float, float]


class NetworkProtocol:
    CALLSIGN_LEN = 32
    MOTTO_LEN = 128
    SERVER_LEN = 8
    MESSAGE_LEN = 128
    HASH_LEN = 64
    WORLD_SETTING_SIZE = 30



from typing import List

from bzflag.networking.game_data_player import PlayerData, IsRegistered, IsVerified, IsAdmin
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgPlayerInfoPacket(GamePacket):
    __slots__ = (
        'players',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgPlayerInfo'
        self.players: List[PlayerData] = []

    def _unpack(self):
        count: int = Packet.unpack_uint8(self.buffer)

        for i in range(0, count):
            player = PlayerData()
            player.player_id = Packet.unpack_uint8(self.buffer)

            properties: int = Packet.unpack_uint8(self.buffer)

            player.is_registered = properties & IsRegistered == IsRegistered
            player.is_verified = properties & IsVerified == IsVerified
            player.is_admin = properties & IsAdmin == IsAdmin

            self.players.append(player)



import math
import socket
import struct

from datetime import datetime, timezone
from io import BytesIO, BufferedIOBase
from typing import BinaryIO, Union, Tuple, Optional, Any, List

from bzflag.networking.game_data_firing_info import FiringInfoData
from bzflag.networking.game_data_flag import FlagData
from bzflag.networking.game_data_player_state import PlayerStateData, JumpJets, OnDriver, UserInputs, PlaySound
from bzflag.networking.game_data_shot import ShotData
from bzflag.networking.network_message import NetworkMessage
from bzflag.networking.unpackable import Unpackable
from bzflag.networking.network_protocol import Vector3F


small_scale = 32766.0
small_max_dist = 0.02 * small_scale
small_max_vel = 0.01 * small_scale
small_max_ang_vel = 0.001 * small_scale


class Packet(Unpackable):
    __slots__ = (
        'mode',
        'code',
        'length',
        'next_file_pos',
        'prev_file_pos',
        'timestamp',
        'data',
    )

    def __init__(self):
        self.mode: int = -1
        self.code: int = -1
        self.length: int = -1
        self.next_file_pos: int = -1
        self.prev_file_pos: int = -1
        self.timestamp: Optional[datetime] = None
        self.data: Any = None

    def unpack(self, buf: BinaryIO) -> None:
        chunk = BytesIO(buf.read(32))

        self.mode = Packet.unpack_uint16(chunk)
        self.code = Packet.unpack_uint16(chunk)
        self.length = Packet.unpack_uint32(chunk)
        self.next_file_pos = Packet.unpack_uint32(chunk)
        self.prev_file_pos = Packet.unpack_uint32(chunk)
        self.timestamp = Packet.unpack_timestamp(chunk)

        if self.length > 0:
            self.data = buf.read(self.length)
        else:
            self.data = None

    @staticmethod
    def unpack_int8(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 1)

    @staticmethod
    def unpack_uint8(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 1, False)

    @staticmethod
    def unpack_int16(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 2)

    @staticmethod
    def unpack_uint16(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 2, False)

    @staticmethod
    def unpack_int32(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 4)

    @staticmethod
    def unpack_uint32(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 4, False)

    @staticmethod
    def unpack_int64(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 8)

    @staticmethod
    def unpack_uint64(buf: BinaryIO) -> int:
        return Packet._unpack_int(buf, 8, False)

    @staticmethod
    def unpack_string(buf: BinaryIO, length: int) -> str:
        raw = buf.read(length)
        return raw.strip(b'\x00').decode('utf-8')

    @staticmethod
    def unpack_ip_address(buf: BinaryIO) -> str:
        # This byte was reserved for differentiating between IPv4 and IPv6 addresses
        # However, since BZFlag only supports IPv4, this byte is skipped
        buf.read(1)

        # IP addresses are stored in network byte order (aka Little Endian)
        ip_as_int = int.from_bytes(buf.read(4), byteorder='little', signed=False)

        # Output the IP address as little-endian unsigned long
        #   https://docs.python.org/3/library/struct.html#format-strings
        return socket.inet_ntoa(struct.pack('<L', ip_as_int))

    @staticmethod
    def unpack_float(buf: BinaryIO) -> float:
        val: bytes = buf.read(4)

        if not val:
            return 0.0

        return struct.unpack('>f', val)[0]

    @staticmethod
    def unpack_vector(buf: BinaryIO) -> Vector3F:
        return (
            Packet.unpack_float(buf),
            Packet.unpack_float(buf),
            Packet.unpack_float(buf),
        )

    @staticmethod
    def unpack_flag(buf: BinaryIO) -> FlagData:
        flag = FlagData()

        flag.index = Packet.unpack_uint16(buf)
        flag.abbv = Packet.unpack_string(buf, 2)
        flag.status = Packet.unpack_uint16(buf)
        flag.endurance = Packet.unpack_uint16(buf)
        flag.owner = Packet.unpack_uint8(buf)
        flag.position = Packet.unpack_vector(buf)
        flag.launch_pos = Packet.unpack_vector(buf)
        flag.landing_pos = Packet.unpack_vector(buf)
        flag.flight_time = Packet.unpack_float(buf)
        flag.flight_end = Packet.unpack_float(buf)
        flag.initial_velocity = Packet.unpack_float(buf)

        return flag

    @staticmethod
    def unpack_firing_info(buf: BinaryIO) -> FiringInfoData:
        data: FiringInfoData = FiringInfoData()

        data.time_sent = Packet.unpack_float(buf)
        data.shot = Packet.unpack_shot(buf)
        data.flag = Packet.unpack_string(buf, 2)
        data.lifetime = Packet.unpack_float(buf)

        return data

    @staticmethod
    def unpack_shot(buf: BinaryIO) -> ShotData:
        shot = ShotData()

        shot.player_id = Packet.unpack_uint8(buf)
        shot.shot_id = Packet.unpack_uint16(buf)
        shot.position = Packet.unpack_vector(buf)
        shot.velocity = Packet.unpack_vector(buf)
        shot.delta_time = Packet.unpack_float(buf)
        shot.team = Packet.unpack_uint16(buf)

        return shot

    @staticmethod
    def unpack_player_state(buf: BinaryIO, code: int) -> PlayerStateData:
        # TODO see if `in_order` is necessary of if it can be removed safely
        in_order: int = Packet.unpack_uint32(buf)
        in_status: int = Packet.unpack_uint16(buf)

        state = PlayerStateData()

        if code == NetworkMessage.PlayerUpdate:
            state.position = Packet.unpack_vector(buf)
            state.velocity = Packet.unpack_vector(buf)
            state.azimuth = Packet.unpack_float(buf)
            state.angular_velocity = Packet.unpack_float(buf)
        else:
            Int3 = Tuple[int, int, int]

            pos: Int3 = (
                Packet.unpack_int16(buf),
                Packet.unpack_int16(buf),
                Packet.unpack_int16(buf),
            )
            vel: Int3 = (
                Packet.unpack_int16(buf),
                Packet.unpack_int16(buf),
                Packet.unpack_int16(buf),
            )
            azi: int = Packet.unpack_int16(buf)
            ang_vel: int = Packet.unpack_int16(buf)

            position: List[float] = [0, 0, 0]
            velocity: List[float] = [0, 0, 0]

            for i in range(0, 3):
                position[i] = (float(pos[i]) * small_max_dist) / small_scale
                velocity[i] = (float(vel[i]) * small_max_vel) / small_scale

            state.position = (position[0], position[1], position[2])
            state.velocity = (velocity[0], velocity[1], velocity[2])
            state.azimuth = (float(azi) * math.pi) / small_scale
            state.angular_velocity = (float(ang_vel) * small_max_ang_vel) / small_scale

        if in_status & JumpJets != 0:
            jump_jets = Packet.unpack_uint16(buf)
            state.jump_jets_scale = float(jump_jets) / small_scale

        if in_status & OnDriver != 0:
            # TODO this value is being read in as unsigned, but needs to be converted to signed...?
            # TODO fix this...
            state.physics_driver = Packet.unpack_uint32(buf)

        if in_status & UserInputs != 0:
            speed = Packet.unpack_uint16(buf)
            ang_vel = Packet.unpack_uint16(buf)

            state.user_speed = (float(speed) * small_max_vel) / small_scale
            state.user_ang_vel = (float(ang_vel) * small_max_ang_vel) / small_scale

        if in_status & PlaySound != 0:
            state.sounds = Packet.unpack_uint8(buf)

        return state

    @staticmethod
    def unpack_timestamp(buf: BinaryIO) -> datetime:
        msb: int = Packet.unpack_uint32(buf)
        lsb: int = Packet.unpack_uint32(buf)

        ts_raw = (msb << 32) + lsb
        ts_float = ts_raw / 1000000

        return datetime.fromtimestamp(ts_float, timezone.utc)

    @staticmethod
    def _unpack_int(buf: Union[BinaryIO, BufferedIOBase], size: int, signed: bool = True) -> int:
        return int.from_bytes(buf.read(size), byteorder='big', signed=signed)



from typing import Callable, Dict

from bzflag.networking.msg_add_player import MsgAddPlayerPacket
from bzflag.networking.msg_admin_info import MsgAdminInfoPacket
from bzflag.networking.msg_alive import MsgAlivePacket
from bzflag.networking.msg_capture_flag import MsgCaptureFlagPacket
from bzflag.networking.msg_flag_drop import MsgFlagDropPacket
from bzflag.networking.msg_flag_grab import MsgFlagGrabPacket
from bzflag.networking.msg_flag_update import MsgFlagUpdatePacket
from bzflag.networking.msg_game_time import MsgGameTimePacket
from bzflag.networking.msg_gm_update import MsgGMUpdatePacket
from bzflag.networking.msg_killed import MsgKilledPacket
from bzflag.networking.msg_message import MsgMessagePacket
from bzflag.networking.msg_new_rabbit import MsgNewRabbitPacket
from bzflag.networking.msg_null import MsgNullPacket
from bzflag.networking.msg_pause import MsgPausePacket
from bzflag.networking.msg_player_info import MsgPlayerInfoPacket
from bzflag.networking.msg_player_update import MsgPlayerUpdatePacket
from bzflag.networking.msg_remove_player import MsgRemovePlayerPacket
from bzflag.networking.msg_score import MsgScorePacket
from bzflag.networking.msg_score_over import MsgScoreOverPacket
from bzflag.networking.msg_set_var import MsgSetVarPacket
from bzflag.networking.msg_shot_begin import MsgShotBeginPacket
from bzflag.networking.msg_shot_end import MsgShotEndPacket
from bzflag.networking.msg_team_update import MsgTeamUpdatePacket
from bzflag.networking.msg_teleport import MsgTeleportPacket
from bzflag.networking.msg_time_update import MsgTimeUpdatePacket
from bzflag.networking.msg_transfer_flag import MsgTransferFlagPacket
from bzflag.networking.network_message import NetworkMessage


GamePacketMap: Dict[NetworkMessage, Callable] = {
    NetworkMessage.Null: MsgNullPacket.factory,
    NetworkMessage.AddPlayer: MsgAddPlayerPacket.factory,
    NetworkMessage.AdminInfo: MsgAdminInfoPacket.factory,
    NetworkMessage.Alive: MsgAlivePacket.factory,
    NetworkMessage.CaptureFlag: MsgCaptureFlagPacket.factory,
    NetworkMessage.DropFlag: MsgFlagDropPacket.factory,
    NetworkMessage.GrabFlag: MsgFlagGrabPacket.factory,
    NetworkMessage.FlagUpdate: MsgFlagUpdatePacket.factory,
    NetworkMessage.GameTime: MsgGameTimePacket.factory,
    NetworkMessage.GMUpdate: MsgGMUpdatePacket.factory,
    NetworkMessage.Killed: MsgKilledPacket.factory,
    NetworkMessage.Message: MsgMessagePacket.factory,
    NetworkMessage.NewRabbit: MsgNewRabbitPacket.factory,
    NetworkMessage.Pause: MsgPausePacket.factory,
    NetworkMessage.PlayerInfo: MsgPlayerInfoPacket.factory,
    NetworkMessage.PlayerUpdate: MsgPlayerUpdatePacket.factory,
    NetworkMessage.PlayerUpdateSmall: MsgPlayerUpdatePacket.factory,
    NetworkMessage.RemovePlayer: MsgRemovePlayerPacket.factory,
    NetworkMessage.Score: MsgScorePacket.factory,
    NetworkMessage.ScoreOver: MsgScoreOverPacket.factory,
    NetworkMessage.SetVar: MsgSetVarPacket.factory,
    NetworkMessage.ShotBegin: MsgShotBeginPacket.factory,
    NetworkMessage.ShotEnd: MsgShotEndPacket.factory,
    NetworkMessage.TeamUpdate: MsgTeamUpdatePacket.factory,
    NetworkMessage.Teleport: MsgTeleportPacket.factory,
    NetworkMessage.TimeUpdate: MsgTimeUpdatePacket.factory,
    NetworkMessage.TransferFlag: MsgTransferFlagPacket.factory,
}



from datetime import datetime
from io import BytesIO
from typing import Optional, List, Tuple

from bzflag.networking.packet_not_set_error import PacketNotSetError
from bzflag.utilities.json_serializable import JsonSerializable
from bzflag.networking.packet import Packet


class GamePacket(JsonSerializable):
    __slots__ = (
        'packet_type',
        'packet',
        'buffer',
        'timestamp',
        'timestamp_offset',
    )

    def __init__(self):
        super().__init__()

        self.json_ignored: List[str] = ['buffer', 'packet', 'timestamp_offset']
        self.packet_type: str = ''
        self.packet: Optional[Packet] = None
        self.buffer: Optional[BytesIO] = None
        self.timestamp: Optional[datetime] = None
        self.timestamp_offset: Optional[datetime] = None

    def _unpack(self) -> None:
        raise NotImplementedError

    def build(self) -> None:
        if self.packet is None:
            raise PacketNotSetError

        self.buffer = BytesIO(self.packet.data)
        self._unpack()
        self.buffer.close()

        self.timestamp = self.packet.timestamp

    def get_countdown_time(self) -> Tuple[int, int, int]:
        """
        Get representation of the current countdown time this packet occurred at.
        This value represents the amount of time that is left in this recording.

        :return: A tuple with the following values: hours, minutes, seconds
        """
        if self.packet is None:
            raise PacketNotSetError

        if self.timestamp_offset is None or self.packet.timestamp is None:
            return -1, -1, -1

        delta = self.timestamp_offset - self.packet.timestamp

        return delta.seconds // 3600, delta.seconds // 60, delta.seconds % 60

    @classmethod
    def factory(cls):
        return cls()



from bzflag.networking.game_data import GameData


class BZDBSetting(GameData):
    __slots__ = (
        'name',
        'value',
    )

    def __init__(self):
        super().__init__()

        self.name: str = ''
        self.value: str = ''



from bzflag.networking.game_packet import GamePacket
from bzflag.networking.network_protocol import Vector3F
from bzflag.networking.packet import Packet


class MsgAlivePacket(GamePacket):
    __slots__ = (
        'player_id',
        'position',
        'azimuth',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgAlive'
        self.player_id: int = -1
        self.position: Vector3F = [0, 0, 0]
        self.azimuth: float = 0.0

    def _unpack(self):
        self.player_id = Packet.unpack_uint8(self.buffer)
        self.position = Packet.unpack_vector(self.buffer)
        self.azimuth = Packet.unpack_float(self.buffer)



from typing import List

from bzflag.networking.game_data_bzdb_setting import BZDBSetting
from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgSetVarPacket(GamePacket):
    __slots__ = (
        'settings',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgSetVar'
        self.settings: List[BZDBSetting] = []

    def _unpack(self):
        count: int = Packet.unpack_uint16(self.buffer)

        for i in range(0, count):
            setting: BZDBSetting = BZDBSetting()

            name_len: int = Packet.unpack_uint8(self.buffer)
            setting.name = Packet.unpack_string(self.buffer, name_len)

            value_len: int = Packet.unpack_uint8(self.buffer)
            setting.value = Packet.unpack_string(self.buffer, value_len)

            self.settings.append(setting)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='bzflag-networking',
    version='1.0.2',
    author='Vladimir "allejo" Jimenez',
    author_email='me@allejo.io',
    description='A Python library for reading and unpacking BZFlag network packets',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/allejo/bzflag-networking.py',
    packages=setuptools.find_packages(),
    classifiers={
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    },
)



from bzflag.networking.game_packet import GamePacket
from bzflag.networking.packet import Packet


class MsgGameTimePacket(GamePacket):
    __slots__ = (
        'msb',
        'lsb',
    )

    def __init__(self):
        super().__init__()

        self.packet_type: str = 'MsgGameTime'
        self.msb: int = -1
        self.lsb: int = -1

    def _unpack(self):
        self.msb = Packet.unpack_uint32(self.buffer)
        self.lsb = Packet.unpack_uint32(self.buffer)



from bzflag.networking.game_data import GameData


class PlayerInfo(GameData):
    __slots__ = (
        'player_index',
        'ip_address',
    )

    def __init__(self):
        super().__init__()

        self.player_index: int = -1
        self.ip_address: str = ''



from typing import BinaryIO, Optional

from bzflag.utilities.json_serializable import JsonSerializable
from bzflag.networking.unpackable import Unpackable
from bzflag.networking.network_protocol import NetworkProtocol
from bzflag.networking.packet import Packet
from bzflag.replay_duration import ReplayDuration


class ReplayHeader(JsonSerializable, Unpackable):
    __slots__ = (
        'magic_number',
        'version',
        'offset',
        'file_time',
        'player',
        'flags_size',
        'world_size',
        'callsign',
        'motto',
        'server_version',
        'app_version',
        'real_hash',
        'length',
    )

    def __init__(self):
        super().__init__()

        self.magic_number: int = -1
        self.version: int = -1
        self.offset: int = 0
        self.file_time: int = 0
        self.player: int = -1
        self.flags_size: int = 0
        self.world_size: int = 0
        self.callsign: str = ''
        self.motto: str = ''
        self.server_version: str = ''
        self.app_version: str = ''
        self.real_hash: str = ''
        self.length: Optional[ReplayDuration] = None

    def unpack(self, buf: BinaryIO) -> None:
        self.magic_number = Packet.unpack_uint32(buf)
        self.version = Packet.unpack_uint32(buf)
        self.offset = Packet.unpack_uint32(buf)
        self.file_time = Packet.unpack_int64(buf)
        self.player = Packet.unpack_uint32(buf)
        self.flags_size = Packet.unpack_uint32(buf)
        self.world_size = Packet.unpack_uint32(buf)
        self.callsign = Packet.unpack_string(buf, NetworkProtocol.CALLSIGN_LEN)
        self.motto = Packet.unpack_string(buf, NetworkProtocol.MOTTO_LEN)
        self.server_version = Packet.unpack_string(buf, NetworkProtocol.SERVER_LEN)
        self.app_version = Packet.unpack_string(buf, NetworkProtocol.MESSAGE_LEN)
        self.real_hash = Packet.unpack_string(buf, NetworkProtocol.HASH_LEN)

        self.length = ReplayDuration(self.file_time)

        # Skip the appropriate number of bytes since we're not making use of this
        # data yet

        buf.read(4 + NetworkProtocol.WORLD_SETTING_SIZE)

        if self.flags_size > 0:
            buf.read(self.flags_size)

        buf.read(self.world_size)
